class SpliceDescriptor:
    ''' 
    descriptor.SpliceDescriptor handles avail descriptors,
        dtmf descriptors, time descriptors and audio descriptors.
    segmentation.SegmentationDescriptor handles segmentation descriptors.
    '''
    def __repr__(self):
        return str(vars(self))
    
    def chk_identifier(self,bitbin):
        # identiﬁer 32 uimsbf == 0x43554549 (ASCII “CUEI”)
        self.identifier = bitbin.asdecodedhex(32)
        if self.identifier != "CUEI":
            raise ValueError(
                'All descriptors must have a identifier of "CUEI"')
        
    def parse(self,bitbin,tag):
        sd_map = { 0: self.avail_descriptor,
           1: self.dtmf_descriptor,
          # 2: SegmentationDescriptor see three5.segmentation 
           3: self.time_descriptor,
           4: self.audio_descriptor }
        self.tag = tag
        self.chk_identifier(bitbin)
        sd_map[tag](bitbin)

    def avail_descriptor(self,bitbin):
        """
        Table 17 -  avail_descriptor()
        """
        self.name = "Avail Descriptor"
        self.provider_avail_id = bitbin.asint(32)

    def dtmf_descriptor(self,bitbin):
        """
        Table 18 -  DTMF_descriptor()
        """
        self.name = "DTMF Descriptor"
        self.preroll = bitbin.asint(8)
        self.dtmf_count = bitbin.asint(3)
        bitbin.forward(5)
        self.dtmf_chars = []
        for i in range(0, self.dtmf_count):
            self.dtmf_chars.append(bitbin.asint(8))

    def time_descriptor(self,bitbin):
        """
        Table 25 - time_descriptor()
        """
        self.name = "Time Descriptor"
        self.TAI_seconds = bitbin.asint(48)
        self.TAI_ns = bitbin.asint(32)
        self.UTC_offset = bitbin.asint(16)

    def audio_descriptor(self,bitbin):
        """
        Table 26 - audio_descriptor()
        """
        self.name = "Audio Descriptor"
        self.components = []
        self.audio_count = bitbin.asint(4)
        reserved = bitbin.asint(4)
        for i in range(0, self.audio_count):
            comp = {}
            comp["component_tag"] = bitbin.asint(8)
            comp["ISO_code="] = bitbin.asint(24)
            comp["bit_stream_mode"] = bitbin.asint(3)
            comp["num_channels"] = bitbin.asint(4)
            comp["full_srvc_audio"] = bitbin.asflag(1)
            self.components.append(comp)



import json

from base64 import b64decode
from bitn import BitBin
from threefive.command import SpliceCommand
from threefive.descriptor import SpliceDescriptor
from threefive.section import SpliceInfoSection
from threefive.segmentation import SegmentationDescriptor


class Header:
    def __init__(self,packet_data):
        self.pid = self.pts = None
        if 'pid' in packet_data.keys():
            self.pid = packet_data['pid']
        if 'pts' in packet_data.keys():
            self.pts = packet_data['pts']

    def __repr__(self):
        return str(vars(self))
    

class Cue:
    '''
    The threefive.Cue class handles parsing
    SCTE 35 message strings.
    '''
    # splice descriptor tags
    sd_tags = [0,1,2,3,4]
    # splice command types
    cmd_types = [0,4,5,6,7,255] 

    def __init__(self, data, packet_data={}):
        # clear any existing values.
        self.info_section = self.command = False
        self.descriptors = []
        # split off headers, if any.
        payload = self.mk_payload(data)
        # threefive.Stream passes packet_data. 
        self.bitbin = BitBin(payload)
        self.header = Header(packet_data)
        self.info_section = SpliceInfoSection()
        self.info_section.parse(self.bitbin)
        self.set_command()
        self.info_section.descriptor_loop_length = self.bitbin.asint(16)
        self.descriptor_loop()
        self.info_section.crc = hex(int.from_bytes(payload[0:4],
                                            byteorder = 'big'))
                
    def __repr__(self):
        return str(self.get())
    
    def descriptor_loop(self):
        '''
        Cue.descriptor_loop 
        parses all splice descriptors
        '''
        dll = self.info_section.descriptor_loop_length
        while dll > 0:
            try:
                sd = self.set_splice_descriptor()
                sdl = sd.descriptor_length
                dll-= sdl+2
                self.descriptors.append(sd)
            except:
                dll = -1

    def get(self):
        '''
        Returns a dict of 
        the SCTE 35 message data.
        '''
        scte35 = False
        try:
            scte35 = {}
            if self.header.pid !=None:
                scte35['header'] = self.kv_clean(vars(self.header))
            scte35['info_section'] = self.get_info_section()
            scte35['command'] = self.get_command()
            if len(self.descriptors)  > 0:
                scte35['descriptors'] = self.get_descriptors()
        finally:
            return scte35

    def get_command(self):
        return self.kv_clean(vars(self.command))

    def get_descriptors(self):
        return [self.kv_clean(vars(d)) for d in self.descriptors]

    def get_info_section(self):
        return self.kv_clean(vars(self.info_section))       
    
    def kv_clean(self,obj):
        '''
        kv_clean removes items from a dict if the value is None
        '''
        return {k: v for k, v in obj.items() if v is not None}

    def kv_print(self, obj):
        print(json.dumps(obj,indent = 2))

    def mk_bits(self, s):
        '''
        Convert Hex and Base64 strings into bytes.
        '''
        if s[:2].lower() == '0x':
            s = s[2:]
        if s[:2].lower() == 'fc':
            return bytes.fromhex(s)
        try:
            return b64decode(s)
        except:
            return s
        
    def mk_payload(self,data):
        '''
        Cue.mk_payload trims the packet
        header if data is a full SCTE-35 packet
        '''
        if data[0] == 0x47:
            payload = data[5:]
        else:
            payload = self.mk_bits(data)
        return payload
    
    def set_command(self):
        '''
        threefive.Cue.set_command
        checks the command type and if valid,
        the splice command data is parsed.
        '''
        sct = self.info_section.splice_command_type
        if sct not in self.cmd_types:
            raise ValueError('Unknown Splice Command Type')
            return False
        self.command = SpliceCommand()
        self.command.parse(sct,self.bitbin)

    def set_splice_descriptor(self):
        '''

        threefive.Cue.set_splice_descriptor
        is called by threefive.Cue.descriptorloop.
        '''
        # splice_descriptor_tag 8 uimsbf
        tag = self.bitbin.asint(8)
        desc_len = self.bitbin.asint(8)
        if tag in self.sd_tags:
            if tag == 2:
                sd = SegmentationDescriptor()
            else:
                sd = SpliceDescriptor()
            sd.parse(self.bitbin,tag)
            sd.descriptor_length = desc_len
            return sd
        else:
            return False

    def show(self):
        '''
        pretty prints the SCTE 35 message
        '''
        self.kv_print(self.get())

    def show_command(self):
        '''
        pretty prints SCTE 35 splice command
        '''
        self.kv_print(self.get_command())

    def show_descriptors(self):
        '''
        pretty prints SCTE 35 splice descriptors
        '''
        self.kv_print(self.get_descriptors())

    def show_info_section(self):
        '''
        pretty prints SCTE 35 splice info section
        '''
        self.kv_print(self.get_info_section())



import sys
from bitn import BitBin
from .cue import Cue
from functools import partial


class Stream:
    '''
    threefive.Stream(tsdata, show_null = False)
    Fast parse mpegts files and streams
    for SCTE 35 packets
    
    tsdata should be a _io.BufferedReader instance:
    example:
        import threefive
        with open('vid.ts','rb') as tsdata:
            threefive.Stream(tsdata)
            
    SCTE-35 Splice null packets are ignored by default. 
    Set show_null = True to show splice null. 

    '''
    no_pts_stream_ids = [188, 190, 191, 240, 241, 242, 248]
    cmd_types = [4,5,6,7,255] # splice command types
    packet_size = 188
    
    def __init__(self, tsdata, show_null = False):
        # set show_null to parse splice null packets
        if show_null:
            self.cmd_types.append(0)
        self.tsdata = tsdata
        self.PTS = False
        self.until_found = False

    def decode(self):
        '''
        StreamPlus.decode() reads MPEG-TS
        to find SCTE-35 packets.
        ''' 
        for packet in iter( partial(self.tsdata.read, self.packet_size), b''):
            # parse all packets headers first
            self.parse_header(packet) 
            if self.chk_magic(packet):
                cuep = Cue(packet,self.packet_data)                            
                if self.until_found:
                    return cuep
                cuep.show()

    def decode_until_found(self):
        '''
        Stream.decode_until_found() reads MPEG-TS
        to find a SCTE-35 packet and returns the packet
        when found.
        ''' 
        self.until_found = True
        cuep = self.decode()
        if cuep:
            return cuep
        return False

    def proxy(self,func = None):
        '''
        StreamProxy.decode() reads MPEG-TS
        writes all packets to sys.stdout.
        writes scte35 data to sys.stderr.
        The optional func arg allows a function
        to be used for custom handling of the SCTE-35
        cue instance.
        the function should match the interface
            func(cuep)
        Where cuep is an instance of threefive.Cue
        If func is not set, threefive.Cue.show() is called.
        '''
        for packet in iter( partial(self.tsdata.read, self.packet_size), b''):
            # Write every packet to stdout
            sys.stdout.buffer.write(packet)
            self.parse_header(packet) 
            if self.chk_magic(packet):
                cuep = Cue(packet,self.packet_data)
                if not func:
                    sys.stderr.buffer.write(cuep.get())
                else:
                    func(cuep)
                    
    def chk_magic(self,packet):
        '''
        Stream.chk_magic(packet)
        does fast SCTE-35 packet detection
        '''
        if packet[0] == 0x47:
            if packet[5] == 0xfc:
                if packet[6] == 48: 
                    if packet[8] == 0:
                        if packet[15] == 255:
                            return packet[18] in self.cmd_types
            
    def parse_header(self,packet):
        '''
        Stream.parse_header(packet)
        reads a MPEG-TS packet header
        for pid and/or pusi.
        '''
        # tei = packet[1] >> 7
        pusi = packet[1] >> 6 & 1
        if pusi:
            pusidata = packet[4:20]
            self.parse_pusi(pusidata)
        # ts_priority = packet[1] >>5 & 0x1
        pid = (packet[1] & 31) << 8
        pid += packet[2]
        # scramble = packet[2] >>6
        # afc = (packet[2] & 48) >> 4
        # count = packet[2] & 15
        self.packet_data = {'pid':pid,'pts':self.PTS}
      
    def parse_pts(self,bitbin):
        '''
        Stream.parse_pts(bitbin)
        This is the process described in the official
        Mpeg-ts specification.
        '''
        a = bitbin.asint(3) << 30
        bitbin.forward(1)          
        b = bitbin.asint(15) << 15
        bitbin.forward(1)          
        c = bitbin.asint(15)
        d = (a+b+c)/90000.0
        # self.PTS is updated when we find a pts.
        self.PTS=round(d,6)
    
    def parse_pusi(self, pusidata):
        '''
        Stream.parse_pusi(pusidata)
        If the pusi data contains these markers,
        we can pull a PTS value..
        '''
        if pusidata[2] == 1: 
            if pusidata[3] not in self.no_pts_stream_ids:
                if (pusidata[6] >> 6) == 2:
                    if (pusidata[7] >> 6) == 2:
                        if (pusidata[9] >> 4) == 2:
                            bitbin = BitBin(pusidata[9:])
                            bitbin.forward(4)
                            self.parse_pts(bitbin)



class SpliceCommand:
    '''
    command.SpliceCommand handles all splice commands.
    '''
    def parse(self,cmd_type,bitbin): 
        '''
        SpliceCommand.parse calls a method 
        depending on the value of cmd_type
        '''
        cmd_map = {0: self.splice_null,
            4: self.splice_schedule,
            5: self.splice_insert,
            6: self.time_signal,
            7: self.bandwidth_reservation,
            255: self.private_command }
        cmd_map[cmd_type](bitbin)

    def __repr__(self):
        return str(vars(self))
    
    def parse_break(self, bitbin):
        self.break_auto_return = bitbin.asflag(1)
        bitbin.forward(6)
        self.break_duration = bitbin.as90k(33)

    def splice_time(self, bitbin):  # 40bits
        self.time_specified_flag = bitbin.asflag(1)
        if self.time_specified_flag:
            bitbin.forward(6)
            self.pts_time = bitbin.as90k(33)
        else:
            bitbin.forward(7)

    def splice_null(self,bitbin):
        """
        Table 7 - splice_null()
        """
        self.name = "Splice Null"
        self.splice_command_length = 0

    def splice_schedule(Self,bitbin):
        """
        Table 8 - splice_schedule()
        """
        self.name = "Splice Schedule"
        splice_count = bitbin.asint(8)
        for i in range(0, splice_count):
            self.splice_event_id = bitbin.asint(32)
            self.splice_event_cancel_indicator = bitbin.asflag(1)
            bitbin.forward(7)
            if not self.splice_event_cancel_indicator:
                self.out_of_network_indicator = bitbin.asflag(1)
                self.program_splice_flag = bitbin.asflag(1)
                self.duration_flag = bitbin.asflag(1)
                bitbin.forward(5)
                if self.program_splice_flag:
                    self.utc_splice_time = bitbin.asint(32)
                else:
                    self.component_count = bitbin.asint(8)
                    self.components = []
                    for j in range(0, self.component_count):
                        self.components[j] = {
                            "component_tag": bitbin.asint(8),
                            "utc_splice_time": bitbin.asint(32),
                        }
                if self.duration_flag: self.break_duration(bitbin)
                self.unique_program_id = bitbin.asint(16)
                self.avail_num = bitbin.asint(8)
                self.avails_expected = bitbin.asint(8)
      
    def splice_insert(self,bitbin):
        """
        Table 9 - splice_insert()
        """
        self.name = "Splice Insert"
        self.splice_event_id = bitbin.asint(32) # uint32
        self.splice_event_cancel_indicator = bitbin.asflag(1)
        bitbin.forward(7) #uint8
        if not self.splice_event_cancel_indicator:
            self.out_of_network_indicator = bitbin.asflag(1)
            self.program_splice_flag = bitbin.asflag(1)
            self.duration_flag = bitbin.asflag(1)
            self.splice_immediate_flag = bitbin.asflag(1)
            bitbin.forward(4) #uint8
            if self.program_splice_flag and not self.splice_immediate_flag:
                self.splice_time(bitbin) # uint8 + uint32
            if not self.program_splice_flag:
                self.component_count = bitbin.asint(8)# uint 8
                self.components = []
                for i in range(0, self.component_count):
                    self.components[i] = bitbin.asint(8)
                if not self.splice_immediate_flag: self.splice_time(bitbin)
            if self.duration_flag: self.parse_break(bitbin)
            self.unique_program_id = bitbin.asint(16)
            self.avail_num = bitbin.asint(8)
            self.avail_expected = bitbin.asint(8)
            
    def time_signal(self, bitbin):
        """
        Table 10 - time_signal()
        """
        self.time_specified_flag = None
        self.pts_time = None
        self.name = "Time Signal"
        self.splice_time(bitbin)

    def bandwidth_reservation(self, bitbin):
        """
        Table 11 - bandwidth_reservation()
        """
        self.name = "Bandwidth Reservation"

    def private_command(self, bitbin):
        """
        Table 12 - private_command()
        """
        self.name = "Private Command"
        self.identifier = bitbin.asint(32)



#!/usr/bin/env python3

import setuptools

with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="threefive",
    version="2.1.81",
    author="fu-corp",
    author_email="spam@futzu.com",
    description="A Fast SCTE 35 Decoder for Mpeg-TS Video, and Base64 or Hex Encoded Messages.", 
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/futzu/threefive",
    packages=setuptools.find_packages(),
    install_requires=["bitn>=0.0.29",],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ],
    python_requires=">=3.6",
    project_urls={
        'Source': 'https://github.com/futzu/SCTE35-threefive',
    },
)



class SpliceInfoSection:
    '''
    three5.SpliceInfoSection is used to parse
    the splice info section of SCTE-35 messages.
    '''

    def parse(self, bitbin):
        self.table_id = bitbin.ashex(8)
        if self.table_id != '0xfc':
            raise ValueError('splice_info_section.table_id should be 0xfc')
        self.section_syntax_indicator = bitbin.asflag(1)
        self.private = bitbin.asflag(1)
        self.reserved = bitbin.ashex(2)
        if self.reserved != '0x3':
            raise ValueError('splice_info_section.reserved should be 0x3')
        self.section_length = bitbin.asint(12)
        self.protocol_version = bitbin.asint(8)
        if self.protocol_version != 0:
            raise ValueError('splice_info_section.protocol_version should be 0')
        self.encrypted_packet = bitbin.asflag(1)
        self.encryption_algorithm = bitbin.asint(6)
        self.pts_adjustment = bitbin.as90k(33)
        self.cw_index = bitbin.ashex(8)
        self.tier = bitbin.ashex(12)
        self.splice_command_length = bitbin.asint(12)
        self.splice_command_type = bitbin.asint(8)
        self.descriptor_loop_length = 0

    def __repr__(self):
        return str(vars(self))
